/*
 * Function hyphenate() takes as arguments a word to be split up
 * by hyphenation and the room available to fit the first part into.
 * If successful in finding a hyphenation point that will permit
 * a small enough first part, the function returns the length of the
 * first part, including one space for the hyphen that will presumably
 * be added.  Otherwise, the original length of the word is returned.
 *
 * The tables were taken from a dump of TeX (initex) after
 * it got done digesting the file hyphen.tex, and consequently this
 * data is derivative from the work of Donald E. Knuth and Frank
 * M. Liang.  Almost all the code for the hyphenation function was
 * taken from TeX, from sections 923-924 of TeX: The Program, by
 * Donald E. Knuth, 1986, Addison Wesley.  See section 919-920 of
 * that book for other references.
 *
 * This program code is in the public domain, so far as I'm concerned,
 * but it is derivative from TeX, as just noted.
 * Greg Lee, 4/10/92.
 */

#include "fm.h"

#define DEL 127

#include "isoprint.h"
#include "trie_link.h"
#include "trie_char.h"
#include "trie_op.h"

extern int ipa_width[256];
extern int revtab[][4];

static unsigned char hyf_distance[] = { 0,
 0, 3, 1, 1, 2, 1, 2, 2, 0, 1, 3, 0, 0, 3, 1, 0, 1, 1, 4, 4,
 4, 2, 2, 3, 3, 3, 4, 3, 4, 2, 5, 5, 4, 5, 3, 2, 5, 2, 6, 2,
 2, 3, 4, 1, 3, 2, 2, 3, 3, 3, 2, 3, 1, 5, 5, 1, 4, 3, 1, 2,
 2, 3, 4, 3, 4, 4, 3, 2, 3, 1, 2, 7, 5, 1, 2, 2, 3, 5, 2, 2,
 3, 1, 3, 1, 3, 4, 4, 3, 3, 3, 2, 2, 2, 2, 3, 3, 1, 1, 4, 3,
 3, 2, 3, 6, 2, 3, 4, 4, 3, 4, 1, 4, 4, 3, 2, 2, 2, 3, 4, 4,
 3, 4, 4, 2, 3, 2, 2, 4, 4, 2, 2, 4, 3, 4, 3, 3, 4, 4, 3, 3,
 3, 4, 3, 3, 4, 8, 1, 3, 1, 1, 3, 2, 3, 6, 2, 4, 4, 3, 2, 2,
 2, 3, 4, 1, 2, 5, 2, 5, 5, 1, 4, 1, 2, 4, 3, 3, 3, 3, 2, 3,
 2 };

static unsigned char hyf_num[] = { 0,
 4, 4, 1, 3, 5, 5, 3, 4, 5, 4, 5, 2, 3, 5, 3, 1, 2, 5, 5, 5,
 5, 1, 2, 4, 3, 2, 4, 4, 3, 4, 4, 5, 4, 3, 4, 3, 5, 2, 5, 5,
 5, 5, 5, 4, 1, 2, 4, 4, 4, 4, 4, 2, 5, 3, 3, 2, 4, 5, 5, 2,
 4, 2, 2, 5, 4, 4, 3, 2, 2, 4, 1, 5, 4, 3, 4, 5, 3, 4, 1, 3,
 2, 1, 3, 1, 1, 5, 1, 4, 1, 4, 4, 2, 1, 1, 4, 2, 4, 2, 4, 5,
 3, 1, 3, 4, 4, 3, 4, 2, 1, 4, 2, 4, 2, 1, 2, 2, 4, 1, 3, 5,
 3, 4, 5, 5, 2, 5, 4, 3, 4, 2, 4, 2, 5, 5, 2, 4, 4, 4, 2, 1,
 4, 1, 4, 2, 3, 5, 4, 5, 1, 2, 2, 3, 4, 5, 3, 3, 1, 3, 2, 1,
 3, 1, 5, 4, 1, 5, 3, 5, 5, 3, 2, 2, 2, 5, 2, 5, 4, 2, 4, 3,
 5 };

static unsigned char hyf_next[] = { 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 6, 1, 0, 0, 9, 0, 5,
 11, 0, 0, 23, 0, 0, 0, 10, 0, 10, 0, 19, 11, 0, 22, 4, 0, 17, 0, 1,
 6, 10, 42, 12, 0, 3, 3, 17, 6, 1, 4, 7, 1, 25, 5, 12, 45, 1, 12, 59,
 6, 5, 10, 8, 7, 22, 12, 6, 22, 1, 70, 0, 29, 12, 74, 10, 8, 19, 17, 10,
 80, 1, 10, 12, 51, 7, 0, 4, 3, 5, 82, 84, 3, 10, 7, 23, 16, 1, 25, 9,
 13, 12, 7, 0, 1, 4, 106, 0, 1, 23, 9, 8, 25, 4, 9, 4, 84, 10, 2, 2,
 30, 103, 6, 4, 124, 9, 12, 10, 10, 1, 53, 8, 23, 133, 3, 3, 6, 26, 41, 23,
 30, 2, 8, 1, 25, 0, 9, 5, 13, 16, 79, 17, 152, 148, 13, 7, 45, 3, 74, 56,
 9, 8, 10, 13, 1, 11, 1, 25, 10, 13, 11, 13, 12, 4, 17, 4, 79, 94, 17, 1,
 17 };

#define WORDMAX 260

static unsigned char hc[WORDMAX];
static unsigned char hyf[WORDMAX];
static unsigned char asciiword[WORDMAX];
static int asciiindex[WORDMAX];

int hyphenate(unsigned char *word, unsigned char *aword, int plen, int hn, int room)
{
	/*unsigned char *hc, *hyf, *asciiword;*/
	int i, j, z, k, l, v, ret, len = plen;
	/*int hn = strlen(word);*/

	if (room >= hn) return(len);
	if (hn < 5) return(len);
	if (hn + 3 > WORDMAX) return(len);
/*
	if ((hc = (unsigned char *)calloc(hn + 3, 1)) == 0) return(hn);
	if ((hyf = (unsigned char *)calloc(hn + 3, 1)) == 0) {
		free(hc);
		return(hn);
	}
	if ((asciiword = (unsigned char *)calloc(hn + 1, 1)) == 0) {
		free(hc);
		free(hyf);
		return(hn);
	}
*/
	memset(hc,'\0', WORDMAX);
	memset(hyf,'\0', WORDMAX);

	hc[0] = hc[hn+1] = DEL;
	hc[hn+2] = 255;
/*
	for (j = 0; word[j]; j++)
		hc[j+1] = (isalpha(word[j]))? tolower(word[j]) - 1 : DEL;
*/
	for (j = 0, k = 0; len > 0 && k < hn; k++, j++, len--) {
		int ch = word[j];
		if (ch == LIT_ESC) {
			j++; len--;
			if (!len) break;
			ch = word[j];
		}
		else if (ch == UNI_ESC) {
			int h, u = -1, uch;
			unsigned int wc;
			j++; len--;
			if (!len) break;
			ch = word[j];
			j++; len--;
			if (!len) break;
			uch = word[j];

	    		wc = (unsigned int)((uch << 8) | ch);
			if (wc < 0x100) ch = (int)wc;
			else {
			    int rt;
			    h = (int)wc;
			    for (rt = 0; ; rt++) {
				if ( (u=revtab[rt][0]) == -1) break;
				if (h == u) break; 
			    }
			    if (u == -1) ch = DEL;
			    else ch = revtab[rt][1];
			}
		}
		else if (is_ipa(aword[j]) && !ipa_width[ch]) continue;
		ch = isoprint[ch];
		asciiword[k] = ch;
		asciiindex[k] = j;
		hc[k+1] = (isalpha(ch))?
			 tolower(ch) - 1 : DEL;
	}

	for (j = 0; j <= hn - 2; j++) {
		z = hc[j]; l = j;
		while (hc[l] == trie_char[z]) {
			if ( (v = trie_op[z]) ) {
				do {
					i = l - hyf_distance[v];
					if (hyf_num[v] > hyf[i]) hyf[i] = hyf_num[v];
					v = hyf_next[v];
				} while (v);
			}
			l++;  z = trie_link[z] + hc[l];
		}
	}

	hyf[1] = 0; hyf[hn-2] = 0; hyf[hn-1] = 0; hyf[hn] = 0;
	for (j = 0; asciiword[j]; j++) {
		if (!isalpha(asciiword[j])) {
			hyf[j+1] = 0;
			hyf[j+2] = 0;
			hyf[j] = 0;
			if (j > 0) hyf[j-1] = 0;
		}
		if (asciiword[j] == '-') hyf[j+1] = 1;
	}

	ret = hn;
	for (j = hn-3; j > 1; j--) {
		z = (hc[j]+1 == '-')? j : j+1;
		if ((hyf[j] % 2) && (z <= room)) {
			ret = z;
			break;
		}
	}
	/*free(hc);*/
	/*free(hyf);*/
	/*return(ret);*/
	ret = asciiindex[ret];
	if (!ret) return plen;
	return(ret);
}
