/*
 * u8tofm.c:
 * Translate unicode utf8 text to ascii with fm
 *   formatting commands.
 * usage: u8tofm -[hdot] <file
 *	-h	use \char"<hex> for unprintable characters
 *	-o	use \char'<octal> for unprintable characters
 *	-d	use \char<decimal> for unprintable characters
 *	-t	put diacritics in front, when possible
 *
 * Version 0.1
 * 			Greg, lee@hawaii.edu, 10/01
 */
#include "fm.h"
#include "revtab.h"

#define TRUE 1
#define FALSE 0

extern char *optarg;		/* from getopt */
extern int  optind;

static char *pre[] = {
	"`",   /* 0 grave */
	"'",   /* 1 acute */
	"^",   /* 2 circumflex */
	"~",   /* 3 tilde */
	"\"",  /* 4 umlaut */
	"H ",  /* 5 double acute */
	"r ",  /* 6 ring over */
	"v ",  /* 7 wedge */
	"u ",  /* 8 breve */
	"=",   /* 9 macron */
	".",   /* 10 dot over */
	"c ",  /* 11 cedilla */
	"k ",  /* 12 ogonek */
	"b ",  /* macron under (9) */
	"d ",  /* dot under (10) */
};

static int d_opt = FALSE,
	h_opt = FALSE,
	o_opt = FALSE,
	t_opt = FALSE,
	u_opt = FALSE,
	num_opt = FALSE,
	globfont = ROMAN;

static void numlook(int c, int f) {
	int i, j, k;
	unsigned char n = 0, uc;
	uc = (unsigned char)c;

	if (d_opt) {
		printf("\\char%d ", uc);
		return;
	}
	if (h_opt) {
		printf("\\char\"%x ", uc);
		return;
	}
	if (o_opt) {
		printf("\\char'%o ", uc);
		return;
	}
	for (i = ' ' + 1 ; i < 127; i++) {
	  for (j = 0; j < 6; j++) {
		if (f == MATH) n = math_other[i][j];
		else if (f == IPA) n = other[i][j];
		else if (f == ROMAN) n = roman_other[i][1];
		else n = i;
		if (n != i && n == uc) break;
	  }
	  if (n != i && n == uc) break;
	}
	if (n == uc) {
		for (k = 0; k < j + 1; k++) putchar('#');
		putchar(i);
	}
	else printf("\\char%d ", uc);
}
static void showsign(int c) {
    switch(c) {
	case 0xa0: printf("~"); break; /* NO-BREAK SPACE */
	case 0xa1: printf("!`"); break; /* INVERTED EXCLAMATION MARK */
	case 0xa2: printf("#%%"); break; /* CENT SIGN */
	case 0xa3: printf("{\\it\\$}"); break; /* POUND SIGN */
	case 0xa4: printf("\\$"); break; /* CURRENCY SIGN */
	case 0xa5: printf("#Y"); break; /* YEN SIGN */
	case 0xa6: printf("#|"); break; /* BROKEN BAR */
	case 0xa7: printf("\\S "); break; /* SECTION SIGN */
	case 0xa8: printf("\\\""); break; /* DIAERESIS */
	case 0xa9: printf("#C"); break; /* COPYRIGHT SIGN */
	case 0xaa: printf("#F"); break; /* FEMININE ORDINAL INDICATOR */
	case 0xab: printf("#<"); break; /* LEFT-POINTING DOUBLE ANGLE QUOTATION MARK */
	case 0xac: printf("\\not "); break; /* NOT SIGN */
	case 0xad: printf("#-"); break; /* SOFT HYPHEN */
	case 0xae: printf("#R"); break; /* REGISTERED SIGN */
	case 0xaf: printf("\\="); break; /* MACRON */
	case 0xb0: printf("\\deg "); break; /* DEGREE SIGN */
	case 0xb1: printf("#+"); break; /* PLUS-MINUS SIGN */
	case 0xb2: printf("^2"); break; /* SUPERSCRIPT TWO */
	case 0xb3: printf("^3"); break; /* SUPERSCRIPT THREE */
	case 0xb4: printf("\\'"); break; /* ACUTE ACCENT */
	case 0xb5: printf("#U"); break; /* MICRO SIGN */
	case 0xb6: printf("\\P "); break; /* PILCROW SIGN */
	case 0xb7: printf("#."); break; /* MIDDLE DOT */
	case 0xb8: printf("\\c "); break; /* CEDILLA */
	case 0xb9: printf("^1"); break; /* SUPERSCRIPT ONE */
	case 0xba: printf("#M"); break; /* MASCULINE ORDINAL INDICATOR */
	case 0xbb: printf("#>"); break; /* RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK */
	case 0xbc: printf("#1"); break; /* VULGAR FRACTION ONE QUARTER */
	case 0xbd: printf("#2"); break; /* VULGAR FRACTION ONE HALF */
	case 0xbe: printf("#3"); break; /* VULGAR FRACTION THREE QUARTERS */
	case 0xbf: printf("?`"); break; /* INVERTED QUESTION MARK */
	case 0xc6: printf("\\AE "); break; /* CAPITAL LETTER AE */
	case 0xd0: printf("#D"); break; /* CAPITAL LETTER ETH */
	case 0xd7: printf("#X"); break; /* MULTIPLICATION SIGN */
	case 0xd8: printf("\\Oslash "); break; /* CAPITAL LETTER O WITH STROKE */
	case 0xde: printf("#T"); break; /* CAPITAL LETTER THORN */
	case 0xdf: printf("\\ss "); break; /* latin small letter sharp */
	case 0xe6: printf("\\ae "); break; /*  */
	case 0xf7: printf("#/"); break; /* divide */
	case 0xfe: printf("#t"); break; /* */
	default: break;
    }
}

int main(int argc, char **argv)
{	int optc, prefixed;
	int ch;
	int h, i, u = -1, m = -1, ipac, diac, font, tfont;
	unsigned char c;
	unsigned int wc;
	int count, spaces = 0;

	while ((optc = getopt (argc, argv, "hdotu#")) != EOF)
		switch (optc) {
			case 'h': h_opt = TRUE; break;
			case 'd': d_opt = TRUE; break;
			case 'o': o_opt = TRUE; break;
			case 't': t_opt = TRUE; break;
			case 'u': u_opt = TRUE; break;
			case '#': num_opt = TRUE; break;
	    		default: 
		   exit(1);
		}


	globfont = ROMAN;

  while ( (ch = getchar()) != EOF ) {

        c = (unsigned char)ch;

	if (c == ' ') {
		spaces++;
		continue;
	}

        if (c < 0x80) {
	    if (globfont == MATH && (c < '0' || c > '9')) {
  		printf("$");
		globfont = ROMAN;
	    }
	    else if (globfont == IPA && (c < 'a' || c > 'z')) {
  		printf("}");
		globfont = ROMAN;
	    }
	    else if (globfont == SUPER || globfont == SUB) {
  		printf("}");
		globfont = ROMAN;
	    }
	    while (spaces > 0) { putchar(' '); spaces--; }
            putchar(c);
	    continue;
        }
        if (c < 0xC0) continue;

        if (c < 0xE0) {
          wc = (unsigned int)(c & 0x1F) << 6;
          count = 1;
          if (c < 0xC2) continue;
        }
        else if (c < 0xF0) {
          wc = (unsigned int)(c & 0x0F) << 12;
          count = 2;
        }
        else if (c < 0xF8) {
          wc = (unsigned int)(c & 0x07) << 18;
          count = 3;
        }
	else if (c < 0xFC) {
          wc = (unsigned int)(c & 0x03) << 24;
          count = 4;
        }
	else if (c < 0xFE) {
          wc = (unsigned int)(c & 0x01) << 30;
          count = 5;
        }
        else continue;

	while (count) {

	  if ( (ch = getchar()) == EOF ) return 0;
          c = (unsigned char) ch ^ 0x80;
          if (!(c < 0x40)) break;
          wc |= (unsigned int) c << (6 * --count);

	}

        if (count) continue;

	while (spaces > 0) { putchar(' '); spaces--; }

	u = -1;

	h = (int)wc;
	m = -1;
	for (i = 0; ; i++) {
		if ( (u=revtab[i][0]) == -1) break;
		font = revtab[i][1];
		if (h == u) m = i; 
		if (h == u && font == globfont) break; 
	}
	if (u == -1 && m != -1) {
		i = m;
		u = h;
	}
	if (u == -1 || u_opt) {
  		if (globfont == MATH) printf("$");
  		else if (globfont == IPA) printf("}");
  		else if (globfont == SUPER) printf("}");
  		else if (globfont == SUB) printf("}");
		globfont = ROMAN;
		if (wc < 0x100) showsign(h);
		else if (wc >= 0x10000) printf("U+%06X", wc);
		else if (u_opt) printf("U+%04X", h);
		else switch(h) {
				case 0x160: printf("\\u S"); break;
				case 0x17E: printf("\\u Z"); break;
				case 0x2013: printf("--"); break;
				case 0x2014: printf("---"); break;
				case 0x2018: printf("`"); break;
				case 0x2019: printf("'"); break;
				case 0x201C: printf("``"); break;
				case 0x201D: printf("''"); break;
				case 0x2026: printf("\\dots "); break;
				default: printf("\\char\"%04X", h);
		}
		continue;
	}

	font = tfont = revtab[i][1];
	if (is_ipa(tfont)) font = IPA;

	ipac = revtab[i][2];
	diac = revtab[i][3];

	if (globfont == ROMAN && font == IPA && diac != -1) {
		switch(diac) {
			case 0: font = ROMAN; break;
			case 1: font = ROMAN; break;
			case 2: font = ROMAN; break;
			case 3: font = ROMAN; break;
			case 4: font = ROMAN; break;
			case 5: font = ROMAN; break;
			case 6: font = ROMAN; break;
			case 7: font = ROMAN; break;
			case 8: font = ROMAN; break;
			case 9: font = ROMAN; break;
			case 10: font = ROMAN; break;
			case 11: font = ROMAN; break;
			case 12: font = ROMAN; break;
			default: break;
		}
	}

	if (font != globfont) {
		if (globfont == MATH) printf("$");
		else if (globfont == IPA && font != IPA) printf("}");
		else if (globfont == SUPER && font != SUPER) printf("}");
		else if (globfont == SUB && font != SUB) printf("}");

		if (font == IPA) printf("{\\ipa ");
		else if (font == MATH) printf("$");
		else if (font == SUPER) printf("^{");
		else if (font == SUB) printf("_{");
		if (font == BOX) globfont = ROMAN;
		else globfont = font;
	}

	if (font == BOX) {
		if (isupper(ipac)) printf("{\\bf");
		switch (ipac) {
		  case 'Q':
		  case 'q': printf("--");
		  case 'X':
		  case 'x': printf("|");
		  case 'L':
		  case 'l': printf("\\lceil ");
		  case 'K':
		  case 'k': printf("\\rceil ");
		  case 'M':
		  case 'm': printf("\\lfloor ");
		  case 'J':
		  case 'j': printf("\\rfloor ");
		  case 'T':
		  case 't': printf("\\vdash ");
		  case 'U':
		  case 'u': printf("\\dashv ");
		  case 'V':
		  case 'v': printf("\\bot ");
		  case 'W':
		  case 'w': printf("\\top ");
		  case 'N':
		  case 'n': printf("\\|");
		  case 'H': printf("\\bx H");
		  case 'h': printf("\\bx h");
		}
		if (isupper(ipac)) printf("}");
		continue;
	}

	if (t_opt && diac >= 0 && diac <= 12 && isalnum(ipac)) {
		int ixdiac = diac;
		if (tfont == IPAUNDER && diac == 9) ixdiac = 13;
		else if (tfont == IPAUNDER && diac == 10) ixdiac = 14;
		prefixed = TRUE;
		if (diac == ixdiac) {
			if (tfont == IPAUNDER) printf("_");
			else if (tfont == IPAOVER) printf("^");
		}
		printf("\\%s", pre[ixdiac]);
	}
	else prefixed = FALSE;

	if (diac == -1) {
		if (tfont == IPASUB) printf("_");
		else if (tfont == IPASUPER) printf("^");

	}

	if (ipac > ' ' && ipac < 0x7f && ipac != '^' && ipac != '_') {
		if (ipac == '$' || ipac == '#' ||
			 ipac == '%' || ipac == '\\' ||
			 ipac == '{' || ipac == '}' )
			printf("\\");
		printf("%c", ipac);
	}
	else numlook(ipac, font);

	if (prefixed) continue;

	if (diac != -1) {
		if (font == IPAUNDER) printf("_");
		else if (font == IPAOVER) printf("^");
		numlook(diac, font);
	}

  }

  if (globfont == MATH) printf("$");
  else if (globfont == IPA) printf("}");

  return 0;
}
